#!/usr/bin/env python3

# -*- coding: utf-8 -*-

# System modules
from csv import reader
from fnmatch import filter as fnmatch_filter
from json import load
import os
import re

# Third party modules
import numpy as np
from pyspec.file.spec import FileSpec
from pyspec.file.tiff import TiffFile


class ScanParser:
    """Partial implementation of a class representing a SPEC scan and
    some of its metadata.

    :param spec_file_name: path to a SPEC file on the CLASSE DAQ
    :type spec_file_name: str
    :param scan_number: the number of a scan in the SPEC file provided
        with `spec_file_name`
    :type scan_number: int
    """

    def __init__(self,
                 spec_file_name:str,
                 scan_number:int):
        """Constructor method"""

        self.spec_file_name = spec_file_name
        self.scan_number = scan_number

        self._scan_path = None
        self._scan_name = None
        self._scan_title = None

        self._spec_scan = None
        self._spec_command = None
        self._spec_macro = None
        self._spec_args = None
        self._spec_scan_npts = None
        self._spec_scan_data = None
        self._spec_positioner_values = None

        self._detector_data_path = None

    def __repr__(self):
        return (f'{self.__class__.__name__}'
                f'({self.spec_file_name}, {self.scan_number}) '
                f'-- {self.spec_command}')

    @property
    def spec_file(self):
        # NB This FileSpec instance is not stored as a private
        # attribute because it cannot be pickled (and therefore could
        # cause problems for parallel code that uses ScanParsers).
        return FileSpec(self.spec_file_name)

    @property
    def scan_path(self):
        if self._scan_path is None:
            self._scan_path = self.get_scan_path()
        return self._scan_path

    @property
    def scan_name(self):
        if self._scan_name is None:
            self._scan_name = self.get_scan_name()
        return self._scan_name

    @property
    def scan_title(self):
        if self._scan_title is None:
            self._scan_title = self.get_scan_title()
        return self._scan_title

    @property
    def spec_scan(self):
        if self._spec_scan is None:
            self._spec_scan = self.get_spec_scan()
        return self._spec_scan

    @property
    def spec_command(self):
        if self._spec_command is None:
            self._spec_command = self.get_spec_command()
        return self._spec_command

    @property
    def spec_macro(self):
        if self._spec_macro is None:
            self._spec_macro = self.get_spec_macro()
        return self._spec_macro

    @property
    def spec_args(self):
        if self._spec_args is None:
            self._spec_args = self.get_spec_args()
        return self._spec_args

    @property
    def spec_scan_npts(self):
        if self._spec_scan_npts is None:
            self._spec_scan_npts = self.get_spec_scan_npts()
        return self._spec_scan_npts

    @property
    def spec_scan_data(self):
        if self._spec_scan_data is None:
            self._spec_scan_data = self.get_spec_scan_data()
        return self._spec_scan_data

    @property
    def spec_positioner_values(self):
        if self._spec_positioner_values is None:
            self._spec_positioner_values = self.get_spec_positioner_values()
        return self._spec_positioner_values

    @property
    def detector_data_path(self):
        if self._detector_data_path is None:
            self._detector_data_path = self.get_detector_data_path()
        return self._detector_data_path

    def get_scan_path(self):
        """Return the name of the directory containining the SPEC file
        for this scan.

        :rtype: str
        """
        return os.path.dirname(self.spec_file_name)

    def get_scan_name(self):
        """Return the name of this SPEC scan (not unique to scans
        within a single spec file).

        :rtype: str
        """
        raise NotImplementedError

    def get_scan_title(self):
        """Return the title of this spec scan (unique to each scan
        within a spec file).

        :rtype: str
        """
        raise NotImplementedError

    def get_spec_scan(self):
        """Return the `pyspec.file.spec.Scan` object parsed from the
        spec file and scan number provided to the constructor.

        :rtype: pyspec.file.spec.Scan
        """
        return self.spec_file.getScanByNumber(self.scan_number)

    def get_spec_command(self):
        """Return the string command of this SPEC scan.

        :rtype: str
        """
        return self.spec_scan.command

    def get_spec_macro(self):
        """Return the macro used in this scan's SPEC command.

        :rtype: str
        """
        return self.spec_command.split()[0]

    def get_spec_args(self):
        """Return a list of the arguments provided to the macro for
        this SPEC scan.

        :rtype: list[str]
        """
        return self.spec_command.split()[1:]

    def get_spec_scan_npts(self):
        """Return the number of points collected in this SPEC scan

        :rtype: int
        """
        raise NotImplementedError

    def get_spec_scan_data(self):
        """Return a dictionary of all the counter data collected by
        this SPEC scan.

        :rtype: dict[str, numpy.ndarray]
        """
        return dict(zip(self.spec_scan.labels, self.spec_scan.data.T))

    def get_spec_positioner_values(self):
        """Return a dictionary of all the SPEC positioner values
        recorded by SPEC just before the scan began.

        :rtype: dict[str,str]
        """
        positioner_values = dict(self.spec_scan.motor_positions)
        names = list(positioner_values.keys())
        mnemonics = self.spec_scan.motors
        if mnemonics is not None:
            for name,mnemonic in zip(names,mnemonics):
                if name != mnemonic:
                    positioner_values[mnemonic] = positioner_values[name]
        return positioner_values

    def get_detector_data_path(self):
        """Return the name of the directory containing detector data
        collected by this scan.

        :rtype: str
        """
        raise NotImplementedError

    def get_detector_data_file(self, detector_prefix, scan_step_index:int):
        """Return the name of the file containing detector data
        collected at a certain step of this scan.

        :param detector_prefix: the prefix used in filenames for the
            detector
        :type detector_prefix: str
        :param scan_step_index: the index of the point in this scan
            whose detector file name should be returned.
        :type scan_step_index: int
        :rtype: str
        """
        raise NotImplementedError

    def get_detector_data(self, detector_prefix, scan_step_index:int):
        """Return the detector data collected at a certain step of
        this scan.

        :param detector_prefix: the prefix used in filenames for the
            detector
        :type detector_prefix: str
        :param scan_step_index: the index of the point in this scan
            whose detector data should be returned.
        :type scan_step_index: int
        :rtype: numpy.ndarray
        """
        raise NotImplementedError

    def get_spec_positioner_value(self, positioner_name):
        """Return the value of a spec positioner recorded before this
        scan began.

        :param positioner_name: the name or mnemonic of a SPEC motor
            whose position should be returned.
        :raises KeyError: if `positioner_name` is not the name or
            mnemonic of a SPEC motor recorded for this scan.
        :raises ValueError: if the recorded string value of the
            positioner in the SPEC file cannot be converted to a
            float.
        :rtype: float
        """
        try:
            positioner_value = self.spec_positioner_values[positioner_name]
            positioner_value = float(positioner_value)
        except KeyError:
            raise KeyError(f'{self.scan_title}: motor {positioner_name} '
                           'not found for this scan')
        except ValueError:
            raise ValueError(f'{self.scan_title}: could not convert value of'
                             f' {positioner_name} to float: '
                             f'{positioner_value}')
        return positioner_value


class FMBScanParser(ScanParser):
    """Partial implementation of a class representing a SPEC scan
    collected at FMB.
    """

    def get_scan_name(self):
        return os.path.basename(self.spec_file.abspath)

    def get_scan_title(self):
        return f'{self.scan_name}_{self.scan_number:03d}'


class SMBScanParser(ScanParser):
    """Partial implementation of a class representing a SPEC scan
    collected at SMB or FAST.
    """

    def __init__(self, spec_file_name, scan_number):
        super().__init__(spec_file_name, scan_number)

        self._pars = None
        self._par_file_pattern = f'*-*-{self.scan_name}'

    def get_scan_name(self):
        return os.path.basename(self.scan_path)

    def get_scan_title(self):
        return f'{self.scan_name}_{self.scan_number}'

    @property
    def pars(self):
        if self._pars is None:
            self._pars = self.get_pars()
        return self._pars

    def get_pars(self):
        """Return a dictionary of values recorded in the .par file
        associated with this SPEC scan.

        :rtype: dict[str,object]
        """
        # JSON file holds titles for columns in the par file
        json_files = fnmatch_filter(
            os.listdir(self.scan_path),
            f'{self._par_file_pattern}.json')
        if len(json_files) != 1:
            raise RuntimeError(f'{self.scan_title}: cannot find the '
                               '.json file to decode the .par file')
        with open(os.path.join(self.scan_path, json_files[0])) as json_file:
            par_file_cols = load(json_file)
        try:
            par_col_names = list(par_file_cols.values())
            scann_val_idx = par_col_names.index('SCAN_N')
            scann_col_idx = int(list(par_file_cols.keys())[scann_val_idx])
        except:
            raise RuntimeError(f'{self.scan_title}: cannot find scan pars '
                               'without a "SCAN_N" column in the par file')

        par_files = fnmatch_filter(
            os.listdir(self.scan_path),
            f'{self._par_file_pattern}.par')
        if len(par_files) != 1:
            raise RuntimeError(f'{self.scan_title}: cannot find the .par '
                               'file for this scan directory')
        par_dict = None
        with open(os.path.join(self.scan_path, par_files[0])) as par_file:
            par_reader = reader(par_file, delimiter=' ')
            for row in par_reader:
                if len(row) == len(par_col_names):
                    row_scann = int(row[scann_col_idx])
                    if row_scann == self.scan_number:
                        par_dict = {}
                        for par_col_idx,par_col_name in par_file_cols.items():
                            # Convert the string par value from the
                            # file to an int or float, if possible.
                            par_value = row[int(par_col_idx)]
                            try:
                                par_value = int(par_value)
                            except ValueError:
                                try:
                                    par_value = float(par_value)
                                except:
                                    pass
                            par_dict[par_col_name] = par_value

        if par_dict is None:
            raise RuntimeError(f'{self.scan_title}: could not find scan pars '
                               'for scan number {self.scan_number}')
        return par_dict

    def get_counter_gain(self, counter_name):
        """Return the gain of a counter as recorded in the comments of
        a scan in a SPEC file converted to nA/V.

        :param counter_name: the name of the counter
        :type counter_name: str
        :rtype: str
        """
        counter_gain = None
        for comment in self.spec_scan.comments:
            match = re.search(
                f'{counter_name} gain: '  # start of counter gain comments
                '(?P<gain_value>\d+) '  # gain numerical value
                '(?P<unit_prefix>[m|u|n])A/V',  # gain units
                comment)
            if match:
                unit_prefix = match['unit_prefix']
                gain_scalar = 1 if unit_prefix == 'n' \
                    else 1e3 if unit_prefix == 'u' else 1e6
                counter_gain = f'{float(match["gain_value"])*gain_scalar} nA/V'

        if counter_gain is None:
            raise RuntimeError(f'{self.scan_title}: could not get gain for '
                               f'counter {counter_name}')
        return counter_gain


class LinearScanParser(ScanParser):
    """Partial implementation of a class representing a typical line
    or mesh scan in SPEC.
    """
    def __init__(self, spec_file_name, scan_number):
        super().__init__(spec_file_name, scan_number)

        self._spec_scan_motor_mnes = None
        self._spec_scan_motor_vals = None
        self._spec_scan_shape = None
        self._spec_scan_dwell = None

    @property
    def spec_scan_motor_mnes(self):
        if self._spec_scan_motor_mnes is None:
            self._spec_scan_motor_mnes = self.get_spec_scan_motor_mnes()
        return self._spec_scan_motor_mnes

    @property
    def spec_scan_motor_vals(self):
        if self._spec_scan_motor_vals is None:
            self._spec_scan_motor_vals = self.get_spec_scan_motor_vals()
        return self._spec_scan_motor_vals

    @property
    def spec_scan_shape(self):
        if self._spec_scan_shape is None:
            self._spec_scan_shape = self.get_spec_scan_shape()
        return self._spec_scan_shape

    @property
    def spec_scan_dwell(self):
        if self._spec_scan_dwell is None:
            self._spec_scan_dwell = self.get_spec_scan_dwell()
        return self._spec_scan_dwell

    def get_spec_scan_motor_mnes(self):
        """Return the mnemonics of the SPEC motor(s) provided to the
        macro for this scan. If there is more than one motor scanned
        (in a "flymesh" scan, for example), the order of motors in the
        returned tuple will go from the fastest moving motor first to
        the slowest moving motor last.

        :rtype: tuple
        """
        raise NotImplementedError

    def get_spec_scan_motor_vals(self):
        """Return the values visited by each of the scanned motors. If
        there is more than one motor scanned (in a "flymesh" scan, for
        example), the order of motor values in the returned tuple will
        go from the fastest moving motor's values first to the slowest
        moving motor's values last.

        :rtype: tuple
        """
        raise NotImplementedError

    def get_spec_scan_shape(self):
        """Return the number of points visited by each of the scanned
        motors. If there is more than one motor scanned (in a
        "flymesh" scan, for example), the order of number of motor
        values in the returned tuple will go from the number of points
        visited by the fastest moving motor first to the the number of
        points visited by the slowest moving motor last.

        :rtype: tuple
        """
        raise NotImplementedError

    def get_spec_scan_dwell(self):
        """Return the dwell time for each point in the scan as it
        appears in the command string.

        :rtype: float
        """
        raise NotImplementedError

    def get_spec_scan_npts(self):
        """Return the number of points collected in this SPEC scan.

        :rtype: int
        """
        return np.prod(self.spec_scan_shape)

    def get_scan_step(self, scan_step_index:int):
        """Return the index of each motor coordinate corresponding to
        the index of a single point in the scan. If there is more than
        one motor scanned (in a "flymesh" scan, for example), the
        order of indices in the returned tuple will go from the index
        of the value of the fastest moving motor first to the index of
        the value of the slowest moving motor last.

        :param scan_step_index: the index of a single point in the
            scan.
        :type scan_step_index: int
        :rtype: tuple
        """
        scan_steps = np.ndindex(self.spec_scan_shape[::-1])
        i = 0
        while i <= scan_step_index:
            scan_step = next(scan_steps)
            i += 1
        return scan_step

    def get_scan_step_index(self, scan_step:tuple):
        """Return the index of a single scan point corresponding to a
        tuple of indices for each scanned motor coordinate.

        :param scan_step: a tuple of the indices of each scanned motor
            coordinate. If there is more than one motor scanned (in a
            "flymesh" scan, for example), the order of indices should
            go from the index of the value of the fastest moving motor
            first to the index of the value of the slowest moving
            motor last.
        :type scan_step: tuple
        :trype: int
        """
        scan_steps = np.ndindex(self.spec_scan_shape[::-1])
        scan_step_found = False
        scan_step_index = -1
        while not scan_step_found:
            next_scan_step = next(scan_steps)
            scan_step_index += 1
            if next_scan_step == scan_step:
                scan_step_found = True
                break
        return scan_step_index


class FMBLinearScanParser(LinearScanParser, FMBScanParser):
    """Partial implementation of a class representing a typical line
    or mesh scan in SPEC collected at FMB.
    """

    def get_spec_scan_motor_mnes(self):
        if self.spec_macro == 'flymesh':
            return (self.spec_args[0], self.spec_args[5])
        if self.spec_macro in ('flyscan', 'ascan'):
            return (self.spec_args[0],)
        if self.spec_macro in ('tseries', 'loopscan'):
            return ('Time',)
        raise RuntimeError(f'{self.scan_title}: cannot determine scan motors '
                           f'for scans of type {self.spec_macro}')

    def get_spec_scan_motor_vals(self):
        if self.spec_macro == 'flymesh':
            fast_mot_vals = np.linspace(float(self.spec_args[1]),
                                        float(self.spec_args[2]),
                                        int(self.spec_args[3])+1)
            slow_mot_vals = np.linspace(float(self.spec_args[6]),
                                        float(self.spec_args[7]),
                                        int(self.spec_args[8])+1)
            return (fast_mot_vals, slow_mot_vals)
        if self.spec_macro in ('flyscan', 'ascan'):
            mot_vals = np.linspace(float(self.spec_args[1]),
                                   float(self.spec_args[2]),
                                   int(self.spec_args[3])+1)
            return (mot_vals,)
        if self.spec_macro in ('tseries', 'loopscan'):
            return self.spec_scan.data[:,0]
        raise RuntimeError(f'{self.scan_title}: cannot determine scan motors '
                           f'for scans of type {self.spec_macro}')

    def get_spec_scan_shape(self):
        if self.spec_macro == 'flymesh':
            fast_mot_npts = int(self.spec_args[3])+1
            slow_mot_npts = int(self.spec_args[8])+1
            return (fast_mot_npts, slow_mot_npts)
        if self.spec_macro in ('flyscan', 'ascan'):
            mot_npts = int(self.spec_args[3])+1
            return (mot_npts,)
        if self.spec_macro in ('tseries', 'loopscan'):
            return len(np.array(self.spec_scan.data[:,0]))
        raise RuntimeError(f'{self.scan_title}: cannot determine scan shape '
                           f'for scans of type {self.spec_macro}')

    def get_spec_scan_dwell(self):
        if self.spec_macro in ('flymesh', 'flyscan', 'ascan'):
            return float(self.spec_args[4])
        if self.spec_macro in ('tseries', 'loopscan'):
            return float(self.spec_args[1])
        raise RuntimeError(f'{self.scan_title}: cannot determine dwell for '
                           f'scans of type {self.spec_macro}')

    def get_detector_data_path(self):
        return os.path.join(self.scan_path, self.scan_title)


class FMBSAXSWAXSScanParser(FMBLinearScanParser):
    """Concrete implementation of a class representing a scan taken
    with the typical SAXS/WAXS setup at FMB.
    """

    def get_scan_title(self):
        return f'{self.scan_name}_{self.scan_number:03d}'

    def get_detector_data_file(self, detector_prefix, scan_step_index:int):
        scan_step = self.get_scan_step(scan_step_index)
        file_indices = [f'{scan_step[i]:03d}'
                        for i in range(len(self.spec_scan_shape))
                        if self.spec_scan_shape[i] != 1]
        if len(file_indices) == 0:
            file_indices = ['000']
        file_name = f'{self.scan_name}_{detector_prefix}_' \
                    f'{self.scan_number:03d}_{"_".join(file_indices)}.tiff'
        file_name_full = os.path.join(self.detector_data_path, file_name)
        if os.path.isfile(file_name_full):
            return file_name_full
        raise RuntimeError(f'{self.scan_title}: could not find detector image '
                           f'file for detector {detector_prefix} scan step '
                           f'({scan_step})')

    def get_detector_data(self, detector_prefix, scan_step_index:int):
        image_file = self.get_detector_data_file(detector_prefix,
                                                 scan_step_index)
        with TiffFile(image_file) as tiff_file:
            image_data = tiff_file.asarray()
        return image_data


class FMBXRFScanParser(FMBLinearScanParser):
    """Concrete implementation of a class representing a scan taken
    with the typical XRF setup at FMB.
    """

    def get_scan_title(self):
        return f'{self.scan_name}_scan{self.scan_number}'

    def get_detector_data_file(self, detector_prefix, scan_step_index:int):
        scan_step = self.get_scan_step(scan_step_index)
        file_name = f'scan{self.scan_number}_{scan_step[0]:03d}.hdf5'
        file_name_full = os.path.join(self.detector_data_path, file_name)
        if os.path.isfile(file_name_full):
            return file_name_full
        raise RuntimeError(f'{self.scan_title}: could not find detector image '
                           f'file for detector {detector_prefix} scan step '
                           f'({scan_step_index})')

    def get_detector_data(self, detector_prefix, scan_step_index:int):
        # Third party modules
        from h5py import File

        detector_file = self.get_detector_data_file(
            detector_prefix, scan_step_index)
        scan_step = self.get_scan_step(scan_step_index)
        with File(detector_file) as h5_file:
            detector_data = \
                h5_file['/entry/instrument/detector/data'][scan_step[0]]
        return detector_data


class SMBLinearScanParser(LinearScanParser, SMBScanParser):
    """Concrete implementation of a class representing a scan taken
    with the typical powder diffraction setup at SMB.
    """

    def get_spec_scan_dwell(self):
        if self.spec_macro in ('flymesh', 'flyscan', 'ascan'):
            return float(self.spec_args[4])
        if self.spec_macro == 'tseries':
            return float(self.spec_args[1])
        if self.spec_macro == 'wbslew_scan':
            return float(self.spec_args[3])
        raise RuntimeError(f'{self.scan_title}: cannot determine dwell time '
                           f'for scans of type {self.spec_macro}')

    def get_spec_scan_motor_mnes(self):
        if self.spec_macro == 'flymesh':
            return (self.spec_args[0], self.spec_args[5])
        if self.spec_macro in ('flyscan', 'ascan'):
            return (self.spec_args[0],)
        if self.spec_macro in ('tseries', 'loopscan'):
            return ('Time',)
        raise RuntimeError(f'{self.scan_title}: cannot determine scan motors '
                           f'for scans of type {self.spec_macro}')

    def get_spec_scan_motor_vals(self):
        if self.spec_macro == 'flymesh':
            fast_mot_vals = np.linspace(float(self.spec_args[1]),
                                        float(self.spec_args[2]),
                                        int(self.spec_args[3])+1)
            slow_mot_vals = np.linspace(float(self.spec_args[6]),
                                        float(self.spec_args[7]),
                                        int(self.spec_args[8])+1)
            return (fast_mot_vals, slow_mot_vals)
        if self.spec_macro in ('flyscan', 'ascan'):
            mot_vals = np.linspace(float(self.spec_args[1]),
                                   float(self.spec_args[2]),
                                   int(self.spec_args[3])+1)
            return (mot_vals,)
        if self.spec_macro in ('tseries', 'loopscan'):
            return self.spec_scan.data[:,0]
        raise RuntimeError(f'{self.scan_title}: cannot determine scan motors '
                           f'for scans of type {self.spec_macro}')

    def get_spec_scan_shape(self):
        if self.spec_macro == 'flymesh':
            fast_mot_npts = int(self.spec_args[3])+1
            slow_mot_npts = int(self.spec_args[8])+1
            return (fast_mot_npts, slow_mot_npts)
        if self.spec_macro in ('flyscan', 'ascan'):
            mot_npts = int(self.spec_args[3])+1
            return (mot_npts,)
        if self.spec_macro in ('tseries', 'loopscan'):
            return len(np.array(self.spec_scan.data[:,0]))
        raise RuntimeError(f'{self.scan_title}: cannot determine scan shape '
                           f'for scans of type {self.spec_macro}')

    def get_spec_scan_dwell(self):
        if self.spec_macro == 'flymesh':
            return float(self.spec_args[4])
        if self.spec_macro in ('flyscan', 'ascan'):
            return float(self.spec_args[-1])
        raise RuntimeError(f'{self.scan_title}: cannot determine dwell time '
                           f'for scans of type {self.spec_macro}')

    def get_detector_data_path(self):
        return os.path.join(self.scan_path, str(self.scan_number))

    def get_detector_data_file(self, detector_prefix, scan_step_index:int):
        scan_step = self.get_scan_step(scan_step_index)
        if len(scan_step) == 1:
            scan_step = (0, *scan_step)
        file_name_pattern = (f'{detector_prefix}_'
                             f'{self.scan_name}_*_'
                             f'{scan_step[0]}_data_'
                             f'{(scan_step[1]+1):06d}.h5')
        file_name_matches = fnmatch_filter(
            os.listdir(self.detector_data_path),
            file_name_pattern)
        if len(file_name_matches) == 1:
            return os.path.join(self.detector_data_path, file_name_matches[0])
        raise RuntimeError(f'{self.scan_title}: could not find detector image '
                           f'file for detector {detector_prefix} scan step '
                           f'({scan_step_index})')

    def get_detector_data(self, detector_prefix, scan_step_index:int):
        # Third party modules
        from h5py import File

        image_file = self.get_detector_data_file(
            detector_prefix, scan_step_index)
        with File(image_file) as h5_file:
            image_data = h5_file['/entry/data/data'][0]
        return image_data


class RotationScanParser(ScanParser):
    """Partial implementation of a class representing a rotation
    scan.
    """

    def __init__(self, spec_file_name, scan_number):
        super().__init__(spec_file_name, scan_number)
        self._starting_image_index = None
        self._starting_image_offset = None

    @property
    def starting_image_index(self):
        if self._starting_image_index is None:
            self._starting_image_index = self.get_starting_image_index()
        return self._starting_image_index

    @property
    def starting_image_offset(self):
        if self._starting_image_offset is None:
            self._starting_image_offset = self.get_starting_image_offset()
        return self._starting_image_offset

    def get_starting_image_index(self):
        """Return the first frame of the detector data collected by
        this scan from the index of the first frame of detector data
        collected by this scan.

        :rtype: int
        """
        raise NotImplementedError

    def get_starting_image_offset(self):
        """Return the offset of the index of the first "good" frame of
        detector data collected by this scan from the index of the
        first frame of detector data collected by this scan.

        :rtype: int
        """
        raise NotImplementedError


class FMBRotationScanParser(RotationScanParser, FMBScanParser):
    """Concrete implementation of a class representing a scan taken
    with the typical tomography setup at FMB.
    """

    def get_spec_scan_npts(self):
        if self.spec_macro == 'flyscan':
            if len(self.spec_args) == 2:
                return 1+int(self.spec_args[0])
            if len(self.spec_args) == 5:
                return 1+int(self.spec_args[3])
            raise RuntimeError(f'{self.scan_title}: cannot obtain number of '
                               f'points from {self.spec_macro} with arguments '
                               f'{self.spec_args}')
        elif self.spec_macro == 'ascan':
            if len(self.spec_args) == 5:
                return int(self.spec_args[3])
            raise RuntimeError(f'{self.scan_title}: cannot obtain number of '
                               f'points from {self.spec_macro} with arguments '
                               f'{self.spec_args}')
        raise RuntimeError(f'{self.scan_title}: cannot determine rotation '
                           f' angles for scans of type {self.spec_macro}')

    def get_starting_image_index(self):
        return 0

    def get_starting_image_offset(self):
        if self.spec_macro == 'flyscan':
#            if len(self.spec_args) == 2:
#                return 1
#            if len(self.spec_args) == 5:
#                return 1
            return 1
#            raise RuntimeError(f'{self.scan_title}: cannot obtain starting '
#                               f'image offset {self.spec_macro} with arguments'
#                               f' {self.spec_args}')
#        elif self.spec_macro == 'ascan':
#            if len(self.spec_args) == 5:
#                return 0
#            raise RuntimeError(f'{self.scan_title}: cannot obtain starting '
#                               f'image offset {self.spec_macro} with arguments'
#                               f' {self.spec_args}')
        raise RuntimeError(f'{self.scan_title}: cannot determine starting '
                           f'image offset for scans of type {self.spec_macro}')

    def get_detector_data_path(self):
        return self.scan_path

    def get_detector_data_file(self, detector_prefix):
        prefix = detector_prefix.upper()
        file_name = f'{self.scan_name}_{prefix}_{self.scan_number:03d}.h5'
        file_name_full = os.path.join(self.detector_data_path, file_name)
        if os.path.isfile(file_name_full):
            return file_name_full
        raise RuntimeError(f'{self.scan_title}: could not find detector image '
                           f'file for detector {detector_prefix}')

    def get_all_detector_data_in_file(
            self, detector_prefix, scan_step_index=None):
        # Third party modules
        from h5py import File

        detector_file = self.get_detector_data_file(detector_prefix)
        with File(detector_file) as h5_file:
            if scan_step_index is None:
                detector_data = h5_file['/entry/instrument/detector/data'][
                    self.starting_image_offset:]
#                sum_det = list(np.sum(detector_data, (1,2)))
#                print(f'\n\nsum scanparser ({len(sum_det)}):\n{sum_det}')
            elif isinstance(scan_step_index, int):
                detector_data = h5_file['/entry/instrument/detector/data'][
                    self.starting_image_offset+scan_step_index]
            elif (isinstance(scan_step_index, (list, tuple))
                    and len(scan_step_index) == 2):
                detector_data = h5_file['/entry/instrument/detector/data'][
                    self.starting_image_offset+scan_step_index[0]:
                    self.starting_image_offset+scan_step_index[1]]
            else:
                raise ValueError('Invalid parameter scan_step_index '
                                 f'({scan_step_index})')
        return detector_data

    def get_detector_data(self, detector_prefix, scan_step_index=None):
        try:
            # Detector files in h5 format
#            print('data in h5 file')
            detector_data = self.get_all_detector_data_in_file(
                detector_prefix, scan_step_index)
#            print(f'detector_data {detector_prefix} {scan_step_index}:\n{detector_data.shape}')
        except:
            # Detector files in tiff format
            if scan_step_index is None:
                detector_data = []
                for index in range(self.spec_scan_npts):
                    detector_data.append(
                        self.get_detector_data(detector_prefix, index))
                detector_data = np.asarray(detector_data)
            elif isinstance(scan_step_index, int):
                image_file = self._get_detector_tiff_file(
                    detector_prefix, scan_step_index)
                if image_file is None:
                    detector_data = None
                else:
                    with TiffFile(image_file) as tiff_file:
                        detector_data = tiff_file.asarray()
            elif (isinstance(scan_step_index, (list, tuple))
                    and len(scan_step_index) == 2):
                detector_data = []
                for index in range(scan_step_index[0], scan_step_index[1]):
                    detector_data.append(
                        self.get_detector_data(detector_prefix, index))
                detector_data = np.asarray(detector_data)
            else:
                raise ValueError('Invalid parameter scan_step_index '
                                 f'({scan_step_index})')
        return detector_data

    def _get_detector_tiff_file(self, detector_prefix, scan_step_index):
        file_name_full = f'{self.spec_file_name}_{detector_prefix.upper()}_' \
            f'{self.scan_number:03d}_' \
            f'{self.starting_image_offset+scan_step_index:03d}.tiff'
        if os.path.isfile(file_name_full):
            return file_name_full
        return None
#        raise RuntimeError(f'{self.scan_title}: could not find detector image '
#                           f'file for scan step ({scan_step_index})')


class SMBRotationScanParser(RotationScanParser, SMBScanParser):
    """Concrete implementation of a class representing a scan taken
    with the typical tomography setup at SMB.
    """

    def __init__(self, spec_file_name, scan_number):
        self._scan_type = None
        super().__init__(spec_file_name, scan_number)

        self._katefix = 0  # RV remove when no longer needed
        self._par_file_pattern = f'id*-*tomo*-{self.scan_name}'

    @property
    def scan_type(self):
        if self._scan_type is None:
            self._scan_type = self.get_scan_type()
        return self._scan_type

    def get_spec_scan_data(self):
        spec_scan_data = super().get_spec_scan_data()
        spec_scan_data['theta'] = np.linspace(
            float(self.pars['ome_start_real']),
            float(self.pars['ome_end_real']), int(self.pars['nframes_real']))
        return spec_scan_data

    def get_spec_scan_npts(self):
        return int(self.pars['nframes_real'])

    def get_scan_type(self):
        scan_type = self.pars.get(
            'tomo_type', self.pars.get(
            'tomotype', self.pars.get('scan_type', None)))
        if scan_type is None:
            raise RuntimeError(
                f'{self.scan_title}: cannot determine the scan_type')
        return scan_type

    def get_starting_image_index(self):
        try:
            junkstart = int(self.pars['junkstart'])
            #RV temp fix for error in par files at Kate's beamline
            #Remove this and self._katefix when no longer needed
            file_name = f'nf_{junkstart:06d}.tif'
            file_name_full = os.path.join(self.detector_data_path, file_name)
            if not os.path.isfile(file_name_full):
                self._katefix = min([
                    int(re.findall(r'\d+', f)[0])
                        for f in os.listdir(self.detector_data_path)
                        if re.match(r'nf_\d+\.tif', f)])
            return junkstart
            #return int(self.pars['junkstart'])
        except:
            raise RuntimeError(f'{self.scan_title}: cannot determine first '
                               'detector image index')

    def get_starting_image_offset(self):
        try:
            return (int(self.pars['goodstart'])-self.starting_image_index)
        except:
            raise RuntimeError(f'{self.scan_title}: cannot determine index '
                               'offset of first good detector image')

    def get_detector_data_path(self):
        return os.path.join(self.scan_path, str(self.scan_number), 'nf')

    def get_detector_data_file(self, scan_step_index:int):
        index = self.starting_image_index + self.starting_image_offset \
            + scan_step_index
        file_name = f'nf_{index:06d}.tif'
        file_name_full = os.path.join(self.detector_data_path, file_name)
        if os.path.isfile(file_name_full):
            return file_name_full
        #RV temp fix for error in par files at Kate's beamline
        #Remove this and self._katefix when no longer needed
        index += self._katefix
        file_name = f'nf_{index:06d}.tif'
        file_name_full = os.path.join(self.detector_data_path, file_name)
        if os.path.isfile(file_name_full):
            return file_name_full
        raise RuntimeError(f'{self.scan_title}: could not find detector image '
                           f'file for scan step ({scan_step_index})')

    def get_detector_data(self, detector_prefix, scan_step_index=None):
#        print(f'\n\nin get_detector_data: {detector_prefix} {scan_step_index}')
        if scan_step_index is None:
            detector_data = []
            for index in range(self.spec_scan_npts):
                detector_data.append(
                    self.get_detector_data(detector_prefix, index))
            detector_data = np.asarray(detector_data)
#            print(f'detector_data shape {type(detector_data)} {detector_data.shape}:\n{detector_data}')
        elif isinstance(scan_step_index, int):
            image_file = self.get_detector_data_file(scan_step_index)
            with TiffFile(image_file) as tiff_file:
                detector_data = tiff_file.asarray()
#            print(f'\t{scan_step_index} {image_file} {np.sum(np.asarray(detector_data))}')
        elif (isinstance(scan_step_index, (list, tuple))
                and len(scan_step_index) == 2):
            detector_data = []
            for index in range(scan_step_index[0], scan_step_index[1]):
                detector_data.append(
                    self.get_detector_data(detector_prefix, index))
            detector_data = np.asarray(detector_data)
        else:
            raise ValueError('Invalid parameter scan_step_index '
                             f'({scan_step_index})')
        return detector_data


class MCAScanParser(ScanParser):
    """Partial implementation of a class representing a scan taken
    while collecting SPEC MCA data.
    """

    def __init__(self, spec_file_name, scan_number):
        super().__init__(spec_file_name, scan_number)

        self._detector_num_bins = None

    def get_detector_num_bins(self, detector_prefix):
        """Return the number of bins for the detector with the given
        prefix.

        :param detector_prefix: the detector prefix as used in SPEC
            MCA data files
        :type detector_prefix: str
        :rtype: int
        """
        raise NotImplementedError


class SMBMCAScanParser(MCAScanParser, SMBLinearScanParser):
    """Concrete implementation of a class representing a scan taken
    with the typical EDD setup at SMB or FAST.
    """

    def get_detector_num_bins(self, detector_prefix):
        with open(self.get_detector_data_file(detector_prefix)) \
             as detector_file:
            lines = detector_file.readlines()
        for line in lines:
            if line.startswith('#@CHANN'):
                try:
                    line_prefix, number_saved, first_saved, last_saved, \
                        reduction_coef = line.split()
                    return int(number_saved)
                except:
                    continue
        raise RuntimeError(f'{self.scan_title}: could not find num_bins for '
                           f'detector {detector_prefix}')

    def get_detector_data_path(self):
        return self.scan_path

    def get_detector_data_file(self, detector_prefix, scan_step_index=0):
        file_name = f'spec.log.scan{self.scan_number}.mca1.mca'
        file_name_full = os.path.join(self.detector_data_path, file_name)
        if os.path.isfile(file_name_full):
            return file_name_full
        raise RuntimeError(
            f'{self.scan_title}: could not find detector image file')

    def get_all_detector_data(self, detector_prefix):
        # This should be easy with pyspec, but there are bugs in
        # pyspec for MCA data.....  or is the 'bug' from a nonstandard
        # implementation of some macro on our end?  According to spec
        # manual and pyspec code, mca data should always begin w/ '@A'
        # In example scans, it begins with '@mca1' instead
        data = []

        with open(self.get_detector_data_file(detector_prefix)) \
                as detector_file:
            lines = [line.strip("\\\n") for line in detector_file.readlines()]

        num_bins = self.get_detector_num_bins(detector_prefix)

        counter = 0
        for line in lines:
            a = line.split()

            if len(a) > 0:
                if a[0] == ("@"+detector_prefix):
                    counter = 1
                    spectrum = np.zeros(num_bins)
            if counter == 1:
                b = np.array(a[1:]).astype('uint16')
                spectrum[(counter-1) * 25:((counter-1) * 25 + 25)] = b
                counter = counter + 1
            elif counter > 1 and counter <= (np.floor(num_bins / 25.)):
                b = np.array(a).astype('uint16')
                spectrum[(counter-1) * 25:((counter-1) * 25 + 25)] = b
                counter = counter + 1
            elif counter == (np.ceil(num_bins/25.)):
                b = np.array(a).astype('uint16')
                spectrum[(counter-1) * 25:
                         ((counter-1) * 25 + (np.mod(num_bins, 25)))] = b
                data.append(spectrum)
                counter = 0

        return np.array(data)

    def get_detector_data(self, detector_prefix, scan_step_index:int):
        detector_data = self.get_all_detector_data(detector_prefix)
        return detector_data[scan_step_index]



# third party modules
import numpy as np
import os
from pathlib import PosixPath
from pydantic import (BaseModel,
                      confloat,
                      conint,
                      conlist,
                      constr,
                      DirectoryPath,
                      FilePath,
                      root_validator,
                      validator)
from scipy.interpolate import interp1d
from typing import Literal, Optional, Union

# local modules
from CHAP.common.models.map import MapConfig
from CHAP.utils.material import Material
from CHAP.utils.parfile import ParFile
from CHAP.utils.scanparsers import SMBMCAScanParser as ScanParser


class MCAElementConfig(BaseModel):
    """Class representing metadata required to configure a single MCA
    detector element.

    :ivar detector_name: name of the MCA used with the scan
    :type detector_name: str
    :ivar num_bins: number of channels on the MCA
    :type num_bins: int
    :ivar include_bin_ranges: list of MCA channel index ranges whose
        data should be included after applying a mask
    :type include_bin_ranges: list[list[int]]
    """
    detector_name: constr(strip_whitespace=True, min_length=1) = 'mca1'
    num_bins: Optional[conint(gt=0)]
    include_bin_ranges: Optional[
        conlist(
            min_items=1,
            item_type=conlist(
                item_type=conint(ge=0),
                min_items=2,
                max_items=2))] = None

    @validator('include_bin_ranges', each_item=True)
    def validate_include_bin_range(cls, value, values):
        """Ensure no bin ranges are outside the boundary of the detector"""
        num_bins = values.get('num_bins')
        if num_bins is not None:
            value[1] = min(value[1], num_bins)
        return value

    def mca_mask(self):
        """Get a boolean mask array to use on this MCA element's data.

        :return: boolean mask array
        :rtype: numpy.ndarray
        """
        mask = np.asarray([False] * self.num_bins)
        bin_indices = np.arange(self.num_bins)
        for min_, max_ in self.include_bin_ranges:
            _mask = np.logical_and(bin_indices > min_, bin_indices < max_)
            mask = np.logical_or(mask, _mask)
        return mask

    def dict(self, *args, **kwargs):
        """Return a representation of this configuration in a
        dictionary that is suitable for dumping to a YAML file.

        :return: dictionary representation of the configuration.
        :rtype: dict
        """
        d = super().dict(*args, **kwargs)
        d['include_bin_ranges'] = [
            list(d['include_bin_ranges'][i]) \
            for i in range(len(d['include_bin_ranges']))]
        return d


class MCAScanDataConfig(BaseModel):
    """Class representing metadata required to locate raw MCA data for
    a single scan and construct a mask for it.

    :ivar spec_file: Path to the SPEC file containing the scan
    :ivar scan_number: Number of the scan in `spec_file`
    :ivar detectors: list of detector element metadata configurations

    :ivar detector_name: name of the MCA used with the scan
    :ivar num_bins: number of channels on the MCA

    :ivar include_bin_ranges: list of MCA channel index ranges whose
        data should be included after applying a mask
    """
    inputdir: Optional[DirectoryPath]
    spec_file: Optional[FilePath]
    scan_number: Optional[conint(gt=0)]
    par_file: Optional[FilePath]
    scan_column: Optional[Union[conint(ge=0), str]]

    detectors: conlist(min_items=1, item_type=MCAElementConfig)

    _parfile: Optional[ParFile]
    _scanparser: Optional[ScanParser]

    class Config:
        underscore_attrs_are_private = False

    @root_validator(pre=True)
    def validate_scan(cls, values):
        """Finalize file paths for spec_file/par_file and flux_file.

        :param values: dictionary of field values to validate
        :type values: dict
        :return: the validated form of `values`
        :rtype: dict
        """
        inputdir = values.get('inputdir')
        spec_file = values.get('spec_file')
        par_file = values.get('par_file')
        flux_file = values.get('flux_file')
        if flux_file:
            if not os.path.isabs(flux_file):
                values['flux_file'] = os.path.join(inputdir, flux_file)
        if spec_file and par_file:
            raise ValueError('Use either spec_file or par_file, not both')
        elif spec_file:
            if inputdir is not None:
                if not os.path.isabs(spec_file):
                    values['spec_file'] = os.path.join(inputdir, spec_file)
        elif par_file:
            if inputdir is not None:
                if not os.path.isabs(par_file):
                    values['par_file'] = os.path.join(inputdir, par_file)
            if 'scan_column' not in values:
                raise ValueError(
                    'When par_file is used, scan_column must be used, too')
            if isinstance(values['scan_column'], str):
                parfile = ParFile(values.get('par_file'))
                if values['scan_column'] not in parfile.column_names:
                    raise ValueError(
                        f'No column named {values["scan_column"]} in '
                        + '{values["par_file"]}. Options: '
                        + ', '.join(parfile.column_names))
        else:
            raise ValueError('Must use either spec_file or par_file')

        return values

    @root_validator
    def validate_detectors(cls, values):
        """Fill in values for _scanparser / _parfile (if
        applicable). Fill in values for each detector's num_bins
        field, if needed.
        """
        if values.get('spec_file', False):
            values['_scanparser'] = ScanParser(values.get('spec_file'),
                                               values.get('scan_number'))
            values['_parfile'] = None
        elif values.get('par_file', False):
            values['_parfile'] = ParFile(values.get('par_file'))
            values['_scanparser'] = ScanParser(
                values['_parfile'].spec_file,
                values['_parfile'].good_scan_numbers()[0])
        for detector in values.get('detectors'):
            if detector.num_bins is None:
                try:
                    detector.num_bins = values['_scanparser']\
                        .get_detector_num_bins(detector.detector_name)
                except Exception as exc:
                    raise ValueError('No value found for num_bins') from exc

        return values

    @property
    def scanparser(self):
        try:
            scanparser = self._scanparser
        except:
            scanparser = ScanParser(self.spec_file, self.scan_number)
            self._scanparser = scanparser
        return scanparser

    def mca_data(self, detector_config, scan_step_index=None):
        """Get the array of MCA data collected by the scan.

        :param detector_config: detector for which data will be returned
        :type detector_config: MCAElementConfig
        :return: MCA data
        :rtype: np.ndarray
        """
        detector_name = detector_config.detector_name
        if self._parfile is not None:
            if scan_step_index is None:
                import numpy as np
                data = np.asarray(
                    [ScanParser(self._parfile.spec_file, scan_number)\
                     .get_all_detector_data(detector_name)[0] \
                     for scan_number in self._parfile.good_scan_numbers()])
            else:
                data = ScanParser(
                    self._parfile.spec_file,
                    self._parfile.good_scan_numbers()[scan_step_index])\
                    .get_all_detector_data(detector_name)
        else:
            if scan_step_index is None:
                data = self.scanparser.get_all_detector_data(
                    detector_name)
            else:
                data = self.scanparser.get_detector_data(
                    detector_config.detector_name, self.scan_step_index)
        return data

    def dict(self, *args, **kwargs):
        """Return a representation of this configuration in a
        dictionary that is suitable for dumping to a YAML file.

        :return: dictionary representation of the configuration.
        :rtype: dict
        """
        d = super().dict(*args, **kwargs)
        for k,v in d.items():
            if isinstance(v, PosixPath):
                d[k] = str(v)
        if d.get('_parfile') is None:
            del d['par_file']
            del d['scan_column']
        else:
            del d['spec_file']
            del d['scan_number']                
        for k in ('_scanparser', '_parfile', 'inputdir'):
            if k in d:
                del d[k]
        return d


class MaterialConfig(BaseModel):
    """Model for parameters to characterize a sample material

    :ivar hexrd_h5_material_file: path to a HEXRD materials.h5 file containing
        an entry for the material properties.
    :ivar hexrd_h5_material_name: Name of the material entry in
        `hexrd_h5_material_file`.
    :ivar lattice_parameter_angstrom: lattice spacing in angstrom to use for
        a cubic crystal.
    """
    material_name: Optional[constr(strip_whitespace=True, min_length=1)]
    lattice_parameters: Optional[Union[
        confloat(gt=0),
        conlist(item_type=confloat(gt=0), min_items=1, max_items=6)]]
    sgnum: Optional[conint(ge=0)]

    _material: Optional[Material]

    class Config:
        underscore_attrs_are_private = False

    @root_validator
    def validate_material(cls, values):
        from CHAP.edd.utils import make_material
        values['_material'] = make_material(values.get('material_name'),
                                            values.get('sgnum'),
                                            values.get('lattice_parameters'))
        return values

    def unique_ds(self, tth_tol=0.15, tth_max=90.0):
        """Get a list of unique HKLs and their lattice spacings

        :param tth_tol: minimum resolvable difference in 2&theta
            between two unique HKL peaks, defaults to `0.15`.
        :type tth_tol: float, optional
        :param tth_max: detector rotation about hutch x axis, defaults
            to `90.0`.
        :type tth_max: float, optional
        :return: unique HKLs and their lattice spacings in angstroms
        :rtype: np.ndarray, np.ndarray
        """
        from CHAP.edd.utils import get_unique_hkls_ds
        return get_unique_hkls_ds([self._material])

    def dict(self, *args, **kwargs):
        """Return a representation of this configuration in a
        dictionary that is suitable for dumping to a YAML file.

        :return: dictionary representation of the configuration.
        :rtype: dict
        """
        d = super().dict(*args, **kwargs)
        for k,v in d.items():
            if isinstance(v, PosixPath):
                d[k] = str(v)
        if '_material' in d:
            del d['_material']
        return d


class MCAElementCalibrationConfig(MCAElementConfig):
    """Class representing metadata & parameters required for
    calibrating a single MCA detector element.

    :ivar max_energy_kev: maximum channel energy of the MCA in keV
    :ivar tth_max: detector rotation about hutch x axis, defaults to `90`.
    :ivar hkl_tth_tol: minimum resolvable difference in 2&theta between two
        unique HKL peaks, defaults to `0.15`.
    :ivar fit_hkls: list of unique HKL indices to fit peaks for in the
        calibration routine
    :ivar tth_initial_guess: initial guess for 2&theta
    :ivar slope_initial_guess: initial guess for detector channel energy
        correction linear slope, defaults to `1.0`.
    :ivar intercept_initial_guess: initial guess for detector channel energy
        correction y-intercept, defaults to `0.0`.
    :ivar tth_calibrated: calibrated value for 2&theta, defaults to None
    :ivar slope_calibrated: calibrated value for detector channel energy
        correction linear slope, defaults to `None`
    :ivar intercept_calibrated: calibrated value for detector channel energy
        correction y-intercept, defaluts to None
    """
    max_energy_kev: confloat(gt=0)
    tth_max: confloat(gt=0, allow_inf_nan=False) = 90.0
    hkl_tth_tol: confloat(gt=0, allow_inf_nan=False) = 0.15
    fit_hkls: Optional[conlist(item_type=conint(ge=0), min_items=1)] = None
    tth_initial_guess: confloat(gt=0, le=tth_max, allow_inf_nan=False)
    slope_initial_guess: float = 1.0
    intercept_initial_guess: float = 0.0
    tth_calibrated: Optional[confloat(gt=0, allow_inf_nan=False)]
    slope_calibrated: Optional[confloat(allow_inf_nan=False)]
    intercept_calibrated: Optional[confloat(allow_inf_nan=False)]

    def fit_ds(self, materials):
        """Get a list of HKLs and their lattice spacings that will be
        fit in the calibration routine

        :return: HKLs to fit and their lattice spacings in angstroms
        :rtype: np.ndarray, np.ndarray
        """
        if not isinstance(materials, list):
            materials = [materials]
        from CHAP.edd.utils import get_unique_hkls_ds
        unique_hkls, unique_ds = get_unique_hkls_ds(materials)

        fit_hkls = np.array([unique_hkls[i] for i in self.fit_hkls])
        fit_ds = np.array([unique_ds[i] for i in self.fit_hkls])

        return fit_hkls, fit_ds


class MCAElementDiffractionVolumeLengthConfig(MCAElementConfig):
    """Class representing input parameters required to perform a
    diffraction volume length measurement for a single MCA detector
    element.

    :ivar measurement_mode: placeholder for recording whether the
        measured DVL value was obtained through the automated
        calculation or a manual selection.
    :type measurement_mode: Literal['manual', 'auto']
    :ivar sigma_to_dvl_factor: to measure the DVL, a gaussian is fit
        to a reduced from of the raster scan MCA data. This variable
        is a scalar that converts the standard deviation of the
        gaussian fit to the measured DVL.
    :type sigma_to_dvl_factor: Optional[Literal[1.75, 1., 2.]]
    :ivar dvl_measured: placeholder for the measured diffraction
        volume length before writing data to file.
    """
    measurement_mode: Optional[Literal['manual', 'auto']] = 'auto'
    sigma_to_dvl_factor: Optional[Literal[3.5, 2., 4.]] = 3.5
    dvl_measured: Optional[confloat(gt=0)] = None

    def dict(self, *args, **kwargs):
        """If measurement_mode is 'manual', exclude
        sigma_to_dvl_factor from the dict representation.
        """
        d = super().dict(*args, **kwargs)
        if self.measurement_mode == 'manual':
            del d['sigma_to_dvl_factor']
        return d


class DiffractionVolumeLengthConfig(MCAScanDataConfig):
    """Class representing metadata required to perform a diffraction
    volume length calculation for an EDD setup using a steel-foil
    raster scan.

    :ivar detectors: list of individual detector elmeent DVL
        measurement configurations
    :type detectors: list[MCAElementDiffractionVolumeLengthConfig]
    """
    detectors: conlist(min_items=1,
                       item_type=MCAElementDiffractionVolumeLengthConfig)

    @property
    def scanned_vals(self):
        """Return the list of values visited by the scanning motor
        over the course of the raster scan.

        :return: list of scanned motor values
        :rtype: np.ndarray
        """
        if self._parfile is not None:
            return self._parfile.get_values(
                self.scan_column,
                scan_numbers=self._parfile.good_scan_numbers())
        return self.scanparser.spec_scan_motor_vals[0]

    @property
    def scanned_dim_lbl(self):
        """Return a label for plot axes corresponding to the scanned
        dimension

        :rtype: str
        """
        if self._parfile is not None:
            return self.scan_column
        return self.scanparser.spec_scan_motor_mnes[0]

class CeriaConfig(MaterialConfig):
    """Model for a Material representing CeO2 used in calibrations.

    :ivar hexrd_h5_material_name: Name of the material entry in
        `hexrd_h5_material_file`, defaults to `'CeO2'`.
    :ivar lattice_parameter_angstrom: lattice spacing in angstrom to use for
        the cubic CeO2 crystal, defaults to `5.41153`.
    """
    material_name: constr(strip_whitespace=True, min_length=1) = 'CeO2'
    sgnum: Optional[conint(ge=0)] = 225
    lattice_parameters: confloat(gt=0) = 5.41153


class MCACeriaCalibrationConfig(MCAScanDataConfig):
    """
    Class representing metadata required to perform a Ceria calibration for an
    MCA detector.

    :ivar scan_step_index: Index of the scan step to use for calibration,
        optional. If not specified, the calibration routine will be performed
        on the average of all MCA spectra for the scan.

    :ivar flux_file: csv file containing station beam energy in eV (column 0)
        and flux (column 1)

    :ivar material: material configuration for Ceria
    :type material: CeriaConfig

    :ivar detectors: list of individual detector element calibration
        configurations
    :type detectors: list[MCAElementCalibrationConfig]

    :ivar max_iter: maximum number of iterations of the calibration routine,
        defaults to `10`.
    :ivar tune_tth_tol: stop iteratively tuning 2&theta when an iteration
        produces a change in the tuned value of 2&theta that is smaller than
        this value, defaults to `1e-8`.
    """
    scan_step_index: Optional[conint(ge=0)]

    flux_file: FilePath

    material: CeriaConfig = CeriaConfig()

    detectors: conlist(min_items=1, item_type=MCAElementCalibrationConfig)

    max_iter: conint(gt=0) = 10
    tune_tth_tol: confloat(ge=0) = 1e-8

    def mca_data(self, detector_config):
        """Get the 1D array of MCA data to use for calibration.

        :param detector_config: detector for which data will be returned
        :type detector_config: MCAElementConfig
        :return: MCA data
        :rtype: np.ndarray
        """
        if self.scan_step_index is None:
            data = super().mca_data(detector_config)
            if self.scanparser.spec_scan_npts > 1:
                data = np.average(data, axis=1)
            else:
                data = data[0]
        else:
            data = super().mca_data(detector_config,
                                    scan_step_index=self.scan_step_index)
        return data

    def flux_correction_interpolation_function(self):
        """
        Get an interpolation function to correct MCA data for relative energy
        flux of the incident beam.

        :return: energy flux correction interpolation function
        :rtype: scipy.interpolate._polyint._Interpolator1D
        """

        flux = np.loadtxt(self.flux_file)
        energies = flux[:,0]/1.e3
        relative_intensities = flux[:,1]/np.max(flux[:,1])
        interpolation_function = interp1d(energies, relative_intensities)
        return interpolation_function


class MCAElementStrainAnalysisConfig(MCAElementConfig):
    """Model for parameters need to perform strain analysis fitting
    for one MCA element.
    """
    tth_max: confloat(gt=0, allow_inf_nan=False) = 90.0
    hkl_tth_tol: confloat(gt=0, allow_inf_nan=False) = 0.15
    fit_hkls: Optional[conlist(item_type=conint(ge=0), min_items=1)] = None
    background: Optional[str]
    peak_models: Union[
        conlist(item_type=Literal['gaussian', 'lorentzian'], min_items=1),
        Literal['gaussian', 'lorentzian']] = 'gaussian'

    tth_file: Optional[FilePath]
    tth_calibrated: Optional[confloat(gt=0, allow_inf_nan=False)]
    slope_calibrated: Optional[confloat(allow_inf_nan=False)]
    intercept_calibrated: Optional[confloat(allow_inf_nan=False)]
    max_energy_kev: Optional[confloat(gt=0)]
    num_bins: Optional[conint(gt=0)]

    _tth_map: Optional[np.ndarray]

    def add_calibration(self, calibration):
        """Finalize values for some fields using a completed
        MCAElementCalibrationConfig that corresponds to the same
        detector.

        :param calibration: MCAElementCalibrationConfig
        :return: None
        """
        add_fields = ['tth_calibrated', 'slope_calibrated',
                      'intercept_calibrated', 'num_bins', 'max_energy_kev']
        for field in add_fields:
            setattr(self, field, getattr(calibration, field))

    def fit_ds(self, materials):
        """Get a list of HKLs and their lattice spacings that will be
        fit in the strain analysis routine

        :return: HKLs to fit and their lattice spacings in angstroms
        :rtype: np.ndarray, np.ndarray
        """
        if not isinstance(materials, list):
            materials = [materials]
        from CHAP.edd.utils import get_unique_hkls_ds
        unique_hkls, unique_ds = get_unique_hkls_ds(materials)

        # unique_hkls, unique_ds = material.unique_ds(
        #     tth_tol=self.hkl_tth_tol, tth_max=self.tth_max)

        fit_hkls = np.array([unique_hkls[i] for i in self.fit_hkls])
        fit_ds = np.array([unique_ds[i] for i in self.fit_hkls])

        return fit_hkls, fit_ds

    def tth_map(self, map_config):
        """Return a map of tth values to use -- may vary at each point
        in the map.

        :param map_config: the map configuration with which the
            returned map of tth values will be used.
        :type map_config: MapConfig
        :return: map of thh values
        :rtype: np.ndarray
        """
        if self._tth_map is not None:
            return self._tth_map
        return np.full(map_config.shape, self.tth_calibrated)


class StrainAnalysisConfig(BaseModel):
    """Model for inputs to CHAP.edd.StrainAnalysisProcessor"""
    inputdir: Optional[DirectoryPath]
    map_config: Optional[MapConfig]
    par_file: Optional[FilePath]
    par_dims: Optional[list[dict[str,str]]]
    other_dims: Optional[list[dict[str,str]]]
    flux_file: FilePath
    detectors: conlist(min_items=1, item_type=MCAElementStrainAnalysisConfig)
    materials: list[MaterialConfig]

    _parfile: Optional[ParFile]

    @root_validator(pre=True)
    def validate_map(cls, values):
        """Ensure exactly one valid map configuration was provided,
        and finalize input filepaths
        """
        inputdir = values.get('inputdir')
        flux_file = values.get('flux_file')
        par_file = values.get('par_file')
        if flux_file:
            if not os.path.isabs(flux_file):
                values['flux_file'] = os.path.join(inputdir, flux_file)
        if par_file:
            if not os.path.isabs(par_file):
                values['par_file'] = os.path.join(inputdir, par_file)
            if 'par_dims' not in values:
                raise ValueError(
                    'If using par_file, must also use par_dims')
            values['_parfile'] = ParFile(values['par_file'])
            values['map_config'] = values['_parfile'].get_map(
                'EDD', 'id1a3', values['par_dims'],
                other_dims=values.get('other_dims', []))
        map_config = values.get('map_config')
        if isinstance(map_config, dict):
            for i, scans in enumerate(map_config.get('spec_scans')):
                spec_file = scans.get('spec_file')
                if not os.path.isabs(spec_file):
                    values['map_config']['spec_scans'][i]['spec_file'] = \
                        os.path.join(inputdir, spec_file)
        return values

    @validator('detectors', pre=True, each_item=True)
    def validate_tth_file(cls, detector, values):
        """Finalize value for tth_file for each detector"""
        inputdir = values.get('inputdir')
        tth_file = detector.get('tth_file')
        if tth_file:
            if not os.path.isabs(tth_file):
                detector['tth_file'] = os.path.join(inputdir, tth_file)
        return detector

    @validator('detectors', each_item=True)
    def validate_tth(cls, detector, values):
        """Validate detector element tth_file field. It may only be
        used if StrainAnalysisConfig used par_file.
        """
        if detector.tth_file is not None:
            if not values.get('par_file'):
                raise ValueError(
                    'variable tth angles may only be used with a '
                    + 'StrainAnalysisConfig that uses par_file.')
            else:
                try:
                    detector._tth_map = ParFile(values['par_file']).map_values(
                        values['map_config'], np.loadtxt(detector.tth_file))
                except Exception as e:
                    raise ValueError(
                        'Could not get map of tth angles from '
                        + f'{detector.tth_file}') from e
        return detector

    def dict(self, *args, **kwargs):
        """Return a representation of this configuration in a
        dictionary that is suitable for dumping to a YAML file.

        :return: dictionary representation of the configuration.
        :rtype: dict
        """
        d = super().dict(*args, **kwargs)
        for k,v in d.items():
            if isinstance(v, PosixPath):
                d[k] = str(v)
        if '_scanparser' in d:
            del d['_scanparser']
        return d

    def mca_data(self, detector_config, map_index):
        """Get MCA data for a single detector element.

        :param detector_config: the detector to get data for
        :type detector_config: MCAElementStrainAnalysisConfig
        :param map_index: index of a single point in the map
        :type map_index: tuple
        :return: one spectrum of MCA data
        :rtype: np.ndarray
        """
        map_coords = {dim: self.map_config.coords[dim][i]
                      for dim,i in zip(self.map_config.dims, map_index)}
        for scans in self.map_config.spec_scans:
            for scan_number in scans.scan_numbers:
                scanparser = scans.get_scanparser(scan_number)
                for scan_step_index in range(scanparser.spec_scan_npts):
                    _coords = {
                        dim.label:dim.get_value(
                            scans, scan_number, scan_step_index)
                        for dim in self.map_config.independent_dimensions}
                    if _coords == map_coords:
                        break
        scanparser = scans.get_scanparser(scan_number)
        return scanparser.get_detector_data(detector_config.detector_name,
                                            scan_step_index)
